//---------------------------------------------------------------------------
// FuncResponseSeedPredation.h
//---------------------------------------------------------------------------
#if !defined(FuncResponseSeedPredation_H)
  #define FuncResponseSeedPredation_H

#include "BehaviorBase.h"

class clGridBase;
/**
 * Models functional response seed predation.  This runs as a
 * model-within-a-model, in weekly timesteps for a year.  It doesn't matter if
 * the length of the overall model timestep is longer than a year; the amount
 * of seeds is treated as a one-year pool, which produces the same result.
 * <p>
 * This behavior reduces the number of seeds available in the "Dispersed Seeds"
 * grid.  In each cell of that grid, there are a certain number of predators
 * (calculated from the initial density of predators) which have the number of
 * seeds present in that cell as a food source.  For all the species to which
 * this behavior is applied, the seeds are eaten with equal enthusiasm; the
 * proportions of each type are captured before predation, and any leftover
 * seeds are redistributed according to those proportions.
 * <p>
 * Seed rain is evenly divided over a set number of timesteps (weeks).  The
 * predator population has as a food source the number of seeds added
 * during the current week's rain (if the rain is going on) plus any leftover
 * seeds from previous weeks which have not been consumed.  Beginning at a
 * certain week in the spring, the number of seeds available to the mice is
 * further reduced by a certain percentage each week to simulate germination.
 * Once germination begins, it continues until the predator model finishes
 * running.  In order to correctly calculate mouse consumption and ensure that
 * the seeds which germinate are actually available later, this keeps track of
 * the seeds consumed; it is this number which is subtracted from total seeds
 * at the end.
 * <p>
 * Seed offtake for each week is calculated as<br>
 * <center>O = <i>IR<sub>s</sub></i> * <i>N</i></center>
 * where
 * <ul>
 * <li><i>0</i> = offtake
 * <li><i>IR<sub>s</sub></i> = per capita seed offtake
 * <li><i>N</i> = number of mice
 * </ul>
 * Per capita seed offtake is<br>
 * <center>IR<sub>s</sub> = <i>c<sub>s</sub></i>(1 - <i>e<sup>-(<i>S*D</i>))
 * </sup></i></center>
 * where
 * <ul>
 * <li><i>c<sub>s</sub></i> = maximum seed intake rate per predator per day
 * (an input parameter)
 * <li><i>S</i> = number of seeds per predator per day
 * <li><i>D</i> = foraging efficiency (an input parameter)
 * </ul>
 * The number of predators in each cell's population is calculated as<br>
 * <center><i>N<sub>t</sub></i> = <i>N<sub>t-1</sub></i> *
 * <i>e<sup>r<sub>t-1</sub></sup></i></center><br>
 * where
 * <ul>
 * <li><i>N<sub>t</sub></i> = number of predators for the current timestep
 * <li><i>N<sub>t-1</sub></i> = number of predators in the previous timestep
 * <li><i>r<sub>t-1</sub></i> = instantaneous rate of change in predator
 * abundance for the previous timestep.
 * </ul>
 * The instantaneous rate of change, r, is calculated as<br>
 * <center><i>r</i> = <i>a</i> + <i>d(IR<sub>s</sub></i>) + <i>g(N)</i></center>
 * where
 * <ul>
 * <li><i>a</i> = the maximum instantaneous rate at which predator abundance
 * delines in the absense of food (an input parameter)
 * <li><i>d</i> = populations demographic efficiency (an input parameter)
 * <li><i>g</i> = coefficient describing the effect that density-dependent
 * factors have on <i>r</i> (an input parameter)
 * <li><i>N</i> = number of predators
 * </ul>
 * <p>
 * This behavior must be used in conjunction with a disperse behavior.  If such
 * a behavior is not present (and thus the "Dispersed Seeds" grid is not
 * present), a fatal error will be thrown during setup.
 * <p>
 * The namestring and parameter file call string are "functional response seed
 * predation".
 * <p>
 * Copyright 2004 Charles D. Canham.
 * @author Lora E. Murphy
 *
  <br>Edit history:
  <br>-----------------
  <br>April 28, 2004 - Submitted as beta (LEM)
 */
class clFuncResponseSeedPredation : public clBehaviorBase {

  public:

  /**
  Constructor.

  @param p_oSimManager Sim Manager object.
  */
  clFuncResponseSeedPredation(clSimManager *p_oSimManager);

  /**
  * Destructor.  Frees memory.
  */
  ~clFuncResponseSeedPredation();

  /**
  * Performs the model.  If predator densities are not to be preserved, then
  * the predator grid is initialized with SetPredatorInitialDensity().
  */
  void Action();

  protected:

  /**
  * Reads in the parameter file values.
  * @param p_oDoc DOM tree of parsed input file.
  * @throw modelErr if there is no such grid called "Dispersed Seeds", or
  * if the "Dispersed Seeds" grid does not contain a valid return code for all
  * behavior species.
  */
  void GetData(xercesc::DOMDocument *p_oDoc);

  /**
  * Sets all grid cells in the predator grid with the appropriate initial
  * density.
  */
  void SetPredatorInitialDensity();

  /**
  * Gets the data from the parameter file.
  * @param p_oDoc DOM tree of parsed input file.
  * @throw modelErr if:
  * <ul>
  * <li>Number of weeks of seedfall is less than or equal to zero
  * <li>Initial predator density is less than or equal to zero
  * <li>Number of weeks to run the model is less than 0 or greater than 52
  * <li>Week to start germination is less than 0 or greater than 52
  * <li>Proportion germinating isn't between 0 and 1
  * </ul>
  */
  void GetParameterFileData(xercesc::DOMDocument *p_oDoc);

  /**Pointer to the "Dispersed Seeds" grid created by disperse behaviors.*/
  clGridBase *mp_oSeedGrid;

    /**Pointer to predator grid.  The name of this grid is "Seed Predators".
    * Its grid cell resolution will match that of "Dispersed Seeds".  It has
    * one float data member - "num preds".  It will accept maps if
    * m_bPreservePredatorDensities = true and if they are of the right
    * resolution.*/
  clGridBase *mp_oPredatorGrid;

  /**File to which to write intermediate output, if desired*/
  char m_cOutput[MAX_FILENAME_SIZE];

    /**Max instantaneous rate at which predator abundance declines in the
    * absense of food, number of predators per week*/
  float m_fMaxInstantaneousDeclineRate;

    /**Population's demographic efficiency*/
  float m_fDemographicEfficiency;

    /**Density-dependent coefficient*/
  float m_fDensityDependentCoefficient;

    /**Foraging efficiency*/
  float m_fForagingEfficiency;

    /**Predator initial density - number per square meter*/
  float m_fPredatorInitialDensity;

    /**Maximum seed intake rate - number of seeds per predator per day*/
  float m_fMaxIntakeRate;

    /**Proportion of seeds that germinate each week in the germination
    * period - this must be a number between 0 and 1*/
  float m_fProportionGerminating;

    /**Area of a seed grid grid cell in square meters - when multiplied by the
    * density of predators, produces the number in the cell*/
  float m_fCellArea;

    /**Number of weeks to run the model*/
  int m_iNumWeeksToModel;

    /**Number of weeks of seed fall*/
  int m_iNumWeeksSeedFall;

    /**Week in which germination begins*/
  int m_iWeekGerminationStarts;

    /**Data member codes for seed grid for number of seeds.  Array size is #
    * behavior species.*/
  short int *mp_iSeedGridCode;

    /**Data member in the predators grid*/
  short int m_iNumPredsCode;

    /**Whether or not to carry over mouse population densities*/
  bool m_bPreservePredatorDensities;
};
//---------------------------------------------------------------------------
#endif // FuncResponseSeedPredation_H
