//---------------------------------------------------------------------------
#include "MichMenGrowthBase.h"
#include "GrowthOrg.h"
#include "ParsingFunctions.h"
#include "SimManager.h"
#include "TreePopulation.h"

//////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
clMichMenBase::clMichMenBase(clSimManager *p_oSimManager) :  
  clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager ), clGrowthBase(p_oSimManager) {
  mp_fSlopeGrowthResponse = NULL;
  mp_fAsympDiamGrowth = NULL;
  mp_fAdultConstRadInc = NULL;
  mp_fAdultConstBAInc = NULL;
  m_bConstBasalAreaLimited = false;
}


//////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
clMichMenBase::~clMichMenBase() {
  if (mp_fSlopeGrowthResponse) {
    delete[] mp_fSlopeGrowthResponse;
    mp_fSlopeGrowthResponse = NULL;
  }

  if (mp_fAsympDiamGrowth) {
    delete[] mp_fAsympDiamGrowth;
    mp_fAsympDiamGrowth = NULL;
  }

  if (mp_fAdultConstRadInc) {
    delete[] mp_fAdultConstRadInc;
    mp_fAdultConstRadInc = NULL;
  }

  if (mp_fAdultConstBAInc) {
    delete[] mp_fAdultConstBAInc;
    mp_fAdultConstBAInc = NULL;
  }
}
//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// GetParameterFileData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
void clMichMenBase::GetParameterFileData(DOMDocument *p_oDoc) {
 try {
   clPopulationBase *p_oTemp = mp_oSimManager->GetPopulationObject("treepopulation");
   clTreePopulation *p_oPop = (clTreePopulation*) p_oTemp;
   DOMElement *p_oElement = p_oDoc->getDocumentElement();
   floatVal *p_fTempValues;
   float fConversionFactor, //for scaling values from per year to per timestep
         fNumberYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();
   short int i; //loop counter

   //Set up the temp array
   p_fTempValues = new floatVal[m_iNumBehaviorSpecies];
   for (i = 0; i < m_iNumBehaviorSpecies; i++)
     p_fTempValues[i].code = mp_iWhatSpecies[i];

   //Fill the values from the parameter file
   //Slope of growth response
   if (mp_fSlopeGrowthResponse) {
     FillSpeciesSpecificValue(p_oElement, "gr_slopeGrowthResponse", "gr_sgrVal",
                          p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true);

     //Transfer to the array
     for (i = 0; i < m_iNumBehaviorSpecies; i++)
       mp_fSlopeGrowthResponse[p_fTempValues[i].code] = p_fTempValues[i].val;
   }

   //Asymptotic diameter growth
   if (mp_fAsympDiamGrowth) {
     FillSpeciesSpecificValue(p_oElement, "gr_asympDiameterGrowth", "gr_adgVal",
                          p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true);

     //Transfer to the array
     for (i = 0; i < m_iNumBehaviorSpecies; i++)
       mp_fAsympDiamGrowth[p_fTempValues[i].code] = p_fTempValues[i].val;
   }

   //Adult constant radial growth increment
   if (mp_fAdultConstRadInc) {
     FillSpeciesSpecificValue(p_oElement, "gr_adultConstRadialInc",
           "gr_acriVal", p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true);

     for (i = 0; i < m_iNumBehaviorSpecies; i++)
       mp_fAdultConstRadInc[p_fTempValues[i].code] = p_fTempValues[i].val;

   }

   //Adult constant area growth increment
   if (mp_fAdultConstBAInc) {

     FillSpeciesSpecificValue(p_oElement, "gr_adultConstAreaInc", "gr_acaiVal",
                          p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true);

     for (i = 0; i < m_iNumBehaviorSpecies; i++)
       mp_fAdultConstBAInc[p_fTempValues[i].code] = p_fTempValues[i].val;
   }

   //Scale the constant radial increment from mm of radial growth per year to
   //cm of diameter growth per timestep.
   //'*2'=>convert radius to dbh; '/10'=>convert mm to cm; *number of years
   //per timestep
   fConversionFactor = 0.2 * fNumberYearsPerTimestep;

   if (mp_fAdultConstRadInc) {
     for (i = 0; i < m_iNumBehaviorSpecies; i++)

       mp_fAdultConstRadInc[mp_iWhatSpecies[i]] *= fConversionFactor;
   }

   //Scale basal area from square cm/year to sqare mm/timestep.
   //*100=>convert from square cm to square mm
   fConversionFactor = 100 * fNumberYearsPerTimestep;

   if (mp_fAdultConstBAInc) {
     for (i = 0; i < m_iNumBehaviorSpecies; i++)

       mp_fAdultConstBAInc[mp_iWhatSpecies[i]] =
              mp_fAdultConstBAInc[mp_iWhatSpecies[i]] *= fConversionFactor;
   }

   //Scale the slope of growth response parameter
   if (mp_fSlopeGrowthResponse && mp_fAsympDiamGrowth) {
     for (i = 0; i < m_iNumBehaviorSpecies; i++)
       if (mp_fSlopeGrowthResponse[mp_iWhatSpecies[i]] != 0.0)
         mp_fSlopeGrowthResponse[mp_iWhatSpecies[i]] =
                 mp_fAsympDiamGrowth[mp_iWhatSpecies[i]] /
                 mp_fSlopeGrowthResponse[mp_iWhatSpecies[i]];
  }

  delete[] p_fTempValues;

 }
 catch (modelErr &err) {throw(err);}
 catch (modelMsg &msg) {throw(msg);} //non-fatal error
 catch (...) {
   modelErr stcErr;
   stcErr.iErrorCode = UNKNOWN;
   strcpy(stcErr.cFunction, "clMichMenBase::GetParameterFileData");
   throw(stcErr);
 }
}


//////////////////////////////////////////////////////////////////////////////
// ApplyGrowthLimits()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
float clMichMenBase::ApplyGrowthLimits(const short int &iSpecies,
  const float &fAmountDiamIncrease, const float &fDiam) {
 try {
   float fGrowthFromIncrement;

   //If growth is limited to either constant radial increment or constant
   //basal increment, take the minimum of that and fAmountDiamIncrease.
   if (m_bConstRadialLimited) {
     fGrowthFromIncrement = (mp_fAdultConstRadInc[iSpecies]);
     if (fGrowthFromIncrement < fAmountDiamIncrease)
       return fGrowthFromIncrement;
       else return fAmountDiamIncrease;
   } else if (m_bConstBasalAreaLimited) {
     fGrowthFromIncrement = mp_fAdultConstBAInc[iSpecies] / fDiam;
     if (fGrowthFromIncrement < fAmountDiamIncrease)
       return fGrowthFromIncrement;
       else return fAmountDiamIncrease;
   }
   else return fAmountDiamIncrease;
 }
 catch (modelErr &err) {throw(err);}
 catch (modelMsg &msg) {throw(msg);} //non-fatal error
 catch (...) {
   modelErr stcErr;
   stcErr.iErrorCode = UNKNOWN;
   strcpy(stcErr.cFunction, "clMichMenBase::ApplyGrowthLimits");
   throw(stcErr);
 }
}
//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// GetGrowthMemberValue()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 21, 2004 - Transformed into clGrowthBase::GetLgm from "AssignLgm"
// June 24, 2005 - Made it so this won't break if light isn't used (LEM)
//////////////////////////////////////////////////////////////////////////////
float clMichMenBase::GetGrowthMemberValue(clTree *p_oTree, float fDiameterGrowth) {

 try {
   float fGrowth, fDiam = 0, fGli = 0;
   int iSp = p_oTree->GetSpecies(), iTp = p_oTree->GetType(),
       iGLICode = mp_oGrowthOrg->GetLightCode(iSp, iTp);

   //Get the tree's diameter and GLI
   p_oTree->GetValue(mp_oGrowthOrg->GetDiamCode(iSp, iTp), &fDiam);
   if (iGLICode > -1) {
     p_oTree->GetValue(mp_oGrowthOrg->GetLightCode(iSp, iTp), &fGli);
   }

   //Prevent older trees that get plenty of light and grow slowly from dying
   //by setting a higher value (if they're basal-area limited)
   if (m_bConstBasalAreaLimited && fDiam > 30.0 && fGli > 10.0)
     fGrowth = 10.0;
   else
     //Otherwise use the amount of growth - convert to radial growth in mm/year
     fGrowth = fDiameterGrowth * m_fConvertCmPerTSToMmPerYr;

   return fGrowth;
 }
 catch (modelErr &err) {throw(err);}
 catch (modelMsg &msg) {throw(msg);} //non-fatal error
 catch (...) {
   modelErr stcErr;
   stcErr.iErrorCode = UNKNOWN;
   strcpy(stcErr.cFunction, "clMichMenBase::GetGrowthMemberValue");
   throw(stcErr);
 }
}
//---------------------------------------------------------------------------
