//---------------------------------------------------------------------------

#ifndef SelectionHarvest
#define SelectionHarvest
//---------------------------------------------------------------------------
#include "BehaviorBase.h"
#define NUM_SIZE_CLASSES 4

class clGridBase;
class clPackage;
class clTree;
class clPopulationBase;

/**
  SelectionHarvest - Version 1.0

  In the case that the parameter file specifies a target basal area, rather than
  an absolute value or percentage value, this behaviour will calculate the
  basal area to be removed based as the difference betweent the target basal area
  the actual basal area (the amount present prior to harvest).(JM)

  This behaviour will be able to set a target basal area for each of the
  4 size classes.  The behaviour will determine the target basal areas in one of two ways.  First, if the user
  specifies their target basal area in all four size classes, the behaviour will
  calculate the difference, for each size class, between the actual basal area and
  the target basal area.  That delta will be the basal area to remove from each
  size class.  The second option is for the user to provide a 'Q-ratio' and a
  total target basal area for the 4 size classes combined.  These values will be
  set in the parameter file. The combination of these two values is used to
  determine the target basal area in each size class based on the exponential
  distribution.  So, regardless of the method, this behaviour will calculate an
  absolute amount of basal area to remove from each size class, based on the
  current basal area in each grid cell.(JM)

  Prior to each harvest event, this behaviour will compute the actual basal
  in each grid cell prior to harvest.  Then, based on the user input target BA values,
  or the Q-ratio method, the behaviour will determine how much basal area must
  be removed from each size class in order to meet the target basal area.(JM)

  If by chance, the current basal area of the system in a size class is less than
  the target basal area, no trees will be harvested from that size class.
  Furthermore, the shortage will not be made up for by overharvesting another
  size class.(JM)

  This behaviour will then create a set of packages for the harvest grids.  Once
  the packages are created, the behaviour will insert them into the proper place
  in the harvest grids.(JM)

  The namestring and parameter file call string for this behavior are both
  "SelectionHarvest". (LEM)

  Copyright 2003 Charles D. Canham.
*/
class clSelectionHarvest : virtual public clBehaviorBase {

 public:

/**
  Constructor.

  @param p_oSimManager Sim Manager object.
*/
 clSelectionHarvest(clSimManager *p_oSimManager);

/**
  Destructor.
*/
 ~clSelectionHarvest();

 /**
 Performs the selection cut setup for a timestep.  For a timestep, this function
 will call CalculateBasalAreaDifference() and TellHarvestToCutTrees() for each
 selection cut for this timestep.
*/
 void Action();

 /**
 * Determines the time that has elapsed since the last harvest for each cell in the
 * harvest results grid.
 */
 void GetTimeSinceHarvest();

 /**
 Types of harvest cuts.  These are listed in increasing order of severity
  (that's important!)
 */
 enum cutType {partial,  /**<Partial cut*/
               gap,      /**<Gap cut*/
               clear};   /**Clear cut*/

 /**
   How amount to cut is defined.
 */
 enum amtType {percentBA,   /**<As a percentage of total basal area in cut range*/
               absBA,       /**<As an amount of BA in sq m in that cut range*/
               percentDen,  /**<As a percentage of total density in that cut range*/
               absDen};     /**<As a total number of trees in that cut range*/


protected:

float  *mp_fLowDBH,  /**<the lower bound for each DBH size class*/
    *mp_fHighDBH, /**<the upper bound for each DBH size class*/
    *mp_fTargetBA,/**<to hold the target basal area for each size class, in m2/ha*/
    *mp_fLandscapeBasalArea, /**<holds the basal area for each size class for all species across the landscape*/
    *mp_fBasalArea,/**<holds the basal area of 1 species for each size class*/
    *mp_fTempTargetBA; /**<used to ensure that TargetBA does not get overwritten in the CalculateBasalAreaDifference function*/
float m_fPlotArea, /**<plot area, in hectares*/
m_fTotalBasalArea[][NUM_SIZE_CLASSES], /**<the array into which GetBasalArea will store the Basal area values for each size class*/
     m_fBasalAreaDifference[][NUM_SIZE_CLASSES];/**<the array into which CalculateBasalAreaDifference will place the absolute basal area to remove from each size class*/


clPackage *mp_oOldPackage,     /**<for finding where to place a new package*/
          *mp_oNewPackage;     /**<a newly created package in the grids*/

clTreePopulation *mp_oPop;   /**<Stashed pointer to tree population*/


/**
HARVEST MASTER CUTS
  <p>
  The grid called "harvestmastercuts" will have a single cell.  This is where the
  cut events are defined.
  <p>
  Each cut event is one package and is applied to a list of species.  It has a
  timestep, cut amount, cut type, and number of cut ranges defined.  It also has
  a unique ID number.  The grid cells to which it is applied are in the
  "harvestcutevents" grid.
  <p>
  For the cut ranges, four dbh ranges can be defined to which the cuts will be
  applied.  The ranges are ordered as 1 is the smallest minimum dbh.  Ranges
  cannot overlap.
  <p>
  The packages are in timestep order, earliest first.
  <p>
  It is possible that a grid map will have been read in for this grid; it will
  be ignored, and any grid created with such a map will be overwritten.
  <p>
  Data members - all for packages:
  <table>
  <tr>
  <th>Data member name</th>      <th>Data type</th>      <th>Description</th>
  </tr>
  <tr>
  <td>id</td>                    <td>int</td>            <td>ID number of cut</td>
  </tr>
  <tr>
  <td>timestep</td>              <td>int</td>            <td>Timestep at which to apply the cut</td>
  </tr>
  <tr>
  <td>cuttype</td>               <td>int</td>            <td>Matches a value of the enum "cutType"</td>
  </tr>
  <tr>
  <td>amttype</td>               <td>int</td>            <td>Matches a value of the enum "amtType"</td>
  </tr>
  <tr>
  <td>species(x)</td>            <td>bool</td>           <td>One of each of these for each species.  If true, this species is being cut.</td>
  </tr>
  <tr>
  <td>rangeamt(x)</td>           <td>float</td>          <td>Range amount to cut.  There are one of each of the following for m_iNumAllowedCutRanges, where x is
  the index, starting at 0 (thus "rangemin0" and "rangeamt2").</td>
  </tr>
  <tr>
  <td>rangemin(x)</td>           <td>float</td>          <td>Range minimum dbh value.  There are one of each of the following for m_iNumAllowedCutRanges, where x is
  the index, starting at 0.</td>
  </tr>
  <tr>
  <td>rangemax(x)</td>           <td>float</td>          <td>Range maximum dbh value.  There are one of each of the following for m_iNumAllowedCutRanges, where x is
  the index, starting at 0.</td>
  </tr>
</table>
*/
clGridBase *mp_oMasterCutsGrid,

/**
HARVEST CUT EVENTS
  <p>
  The grid called "harvestcutevents" will have a cell resolution that matches the
  tree population.  This is where data about harvest cutting events is stored.
  <p>
  The need to cut a grid cell is signaled by the presence of a package.  The
  package has an ID number which matches a package in the "harvestmastercuts" grid,
  which contains the information about how the cut is actually to be performed.
  Packages are in timestep order, earliest first.
  <p>
  It is possible that a grid map will have been read in for this grid; it will
  be ignored, and any grid created with such a map will be overwritten.
  <p>
  Data members - all for packages:
  <p>
  <table>
  <tr>
  <th>Data member name</th><th>Data type</th><th>Description</th>
  </tr>
  <tr>
  <td>id</td>                    <td>int</td>            <td>ID number matching master package in "harvestmastercuts"</td>
  </tr>
  <tr>
  <td>timestep</td>              <td>int</td>            <td>Timestep at which to apply the cut</td>
  </tr>
</table>
*/
*mp_oCutEventsGrid,  /**<"harvestcutevents" grid*/

/**
* RESULTS GRID
* <p>
* If this behavior is harvesting, the grid is called "Harvest Results"; if
* it is episodic mortality, the grid is called "Mortality Episode Results".
* The grid has a cell resolution matching that of "harvestcutevents/
* mortepisodecutevents".  This is where data on actual cut/kill results is
* stored.  The data is stored raw - no conversion to per-hectare amounts.
* <p>
* It is possible that a grid map will have been read in for this grid; it will
* be ignored, and any grid created with such a map will be overwritten.
* <table>
* <tr>
* <th>Data member name</th>      <th>Data type</th>      <th>Description</th>
* </tr>
* <tr>
* <td>Harvest Type</td>           <td>int</td>            <td>Type of harvest that occurred in the current timestep -
*                         -1 if none has occurred.  This data member is not registered if this grid is "Mortality Episode Results".</td>
* </tr>
* <tr>
* <td>Cut Density_x_sp</td>      <td>int</td>            <td>Number of trees cut in the current timestep.  There are one of each of the following for m_iNumAllowedCutRanges times number
* of species, where x is the cut range index, starting at 0, and sp is the
* species number.</td>
* </tr>
* <tr>
* <td>Cut Basal Area_x_sp</td>   <td>float</td>          <td>Total basal area cut in the current timestep.  There are one of each of the following for m_iNumAllowedCutRanges times number
* of species, where x is the cut range index, starting at 0, and sp is the
* species number.</td>
* </tr>
* </table>
*/
*mp_oResultsGrid,

/**
TIME SINCE HARVEST
  <p>
  The grid called "timesinceharvest" will have a cell resolution that matches the
  harvest results grid.  This is where the time since the last harvest is stored.
  <p>
    Data members - all for packages:
  <p>
  <table>
  <tr>
  <th>Data member name</th><th>Data type</th><th>Description</th>
  </tr>
  <tr>
  <td>Time</td>                    <td>int</td>            <td>time since last harvest</td>
  </tr>
</table>
*/

*mp_oTimeSinceHarvestGrid;


//These hold the return codes for the data members
int m_iInitialAge; /**<Initial age*/
           //"harvestmastercuts" grid data members
 short int m_iMasterTimestepCode,   /**<timestep data member code in "harvestmastercuts" grid*/
           m_iMasterIDCode,         /**<id data member code in "harvestmastercuts" grid*/
           *mp_iSpeciesCodes,       /**<species data member code in "harvestmastercuts" grid.  Array size is number species.
                                    Array index matches species number*/
           m_iCutTypeCode,          /**<cuttype data member code in "harvestmastercuts" grid*/
           m_iAmountTypeCode,       /**<amttype data member code in "harvestmastercuts" grid*/
           *mp_iRangeMinCodes,      /**<rangeminx data member codes in "harvestmastercuts" grid.
           Array size is number cut ranges*/
           *mp_iRangeMaxCodes,      /**<rangemaxx data member codes in "harvestmastercuts" grid.
           Array size is number cut ranges*/
           *mp_iRangeAmountCodes,   /**<rangeamtx data member codes in "harvestmastercuts" grid.
           Array size is number cut ranges*/

           //"harvestcutevents" grid data members
           m_iCutTimestepCode,      /**<timestep data member code in "harvestcutevents" grid*/
           m_iCutIDCode,            /**<id data member code in "harvestcutevents" grid*/

           //"Harvest Results" grid data members
           m_iHarvestTypeCode,      /**<Harvest Type data member code in "Harvest Results" grid*/
           **mp_iDenCutCodes,       /**<dencut data member codes in "Harvest Results" grid.
           Array size is number cut ranges by number of species*/
           **mp_iBaCutCodes,        /**<bacut data member codes in "Harvest Results" grid.
           Array size is number cut ranges by number of species*/

 //Other short int variables which are not data member return codes
           m_iNumAllowedCutRanges, /**<Number of cut ranges allowed.*/
           m_iTime, /**<Data member return code for "time" data member of "timesinceharvest" grid*/
           m_iNumSpecies; /**<Total number of species*/

/**
 Reads all the harvest events from the parameter file.  It doesn't
 perform validation on them other than basic data types - more in-depth logical
 validation is left to ValidatePackages().

 @param p_oDoc DOM tree of parsed input file.
*/
 void GetData(xercesc::DOMDocument *p_oDoc);

/**
 Gets the basal area in a cut area for all species.  This gets the basal
 area in each cut range and puts it in an array whose pointer was passed.

 @param p_oTreePop Pointer to the tree population that will have the selection cut applied to it.
 @param p_fTotalBasalArea Array (size m_iNumAllowedCutRanges) in which to put the
        total basal area calculation.
 @param p_fLoDbh Array (size m_iNumAllowedCutRanges) of lower-limit dbhs for the
        cut ranges that have been defined.
 @param p_fHiDbh Array (size m_iNumAllowedCutRanges) of upper-limit dbhs for the
        cut ranges that have been defined.
*/
void GetBasalArea(clTreePopulation *p_oTreePop, float *p_fTotalBasalArea,
      float *p_fLoDbh, float *p_fHiDbh);

/**
 Gets the basal area in a cut area for a single species.  The cut area is
 defined by a linked list of grid cells.  This gets the basal
 area in each cut range and puts it in an array whose pointer was passed.

 @param p_oTreePop Tree population
 @param iSpecies Species of tree for which to calculate basal area.
 @param p_fTotalBasalArea Array (size m_iNumAllowedCutRanges) in which to put the
        total basal area calculation.
 @param p_fLoDbh Array (size m_iNumAllowedCutRanges) of lower-limit dbhs for the
        cut ranges that have been defined.
 @param p_fHiDbh Array (size m_iNumAllowedCutRanges) of upper-limit dbhs for the
        cut ranges that have been defined.
*/
void GetBasalArea(clTreePopulation *p_oTreePop, int iSpecies, float *p_fTotalBasalArea, float *p_fLoDbh, float *p_fHiDbh);

/**
    Determines the dbh to cut in each grid cell by calculating the difference
    between the target basal area and the actual current basal area. The difference
    is stored in an array whose pointer is passed to the function.(JM)

    @param p_fCurrentBasalArea Pointer to array containing the basal area in each
      size class
*/

int CalculateBasalAreaDifference(float *p_fCurrentBasalArea);


/**
  This function will use the values extracted from the parameter file in order
  to create a package which will be inserted into the master cut events grid.
*/
void CreateMasterCutPackage(int iSpecies);

/**
  This function will edit the cut events grid so that it holds ID and timestep
  values for the selection harvest created in the behaviour.  The function will
  traverse through every cell in the cut event grid and create a package
  containing the selection harvest ID and a timestep.
*/

void EditCutEventsGrid();

/**
 Obtains a pointer to the harvest grids and gets all the return codes/
*/
 void GetHarvestGrids();

 /**
 Makes sure that the data in the event harvest packages makes sense.  For
 each cut event (each package) - this checks the following:
 <ul>
 <li>No negative values in cut ranges
 <li>Cut ranges do not overlap
 </ul>
*/
 void ValidatePackages();



};
//---------------------------------------------------------------------------
#endif

